"""Module to read YAML files."""
from pathlib import Path
from typing import Any
from ruamel.yaml import YAML

from trailblazer.constants import FileExtension
from trailblazer.io.validate_path import validate_file_suffix


def read_yaml(file_path: Path) -> Any:
    """Read content in a YAML file."""
    validate_file_suffix(path_to_validate=file_path, target_suffix=FileExtension.YAML)
    yaml: YAML = YAML(typ="safe", pure=True)
    return yaml.load(file_path)



import datetime
from typing import List

import pytest

from tests.apps.tower.conftest import CaseIDs
from tests.mocks.store_mock import MockStore
from trailblazer.constants import TrailblazerStatus
from trailblazer.store.models import Analysis


def test_setup_db(store: MockStore):
    """Test store contains tables."""

    # GIVEN a store which is already setup

    # THEN it should contain tables
    assert store.engine.table_names()


def test_update_analysis_from_slurm_run_status(analysis_store: MockStore, squeue_stream_jobs: str):
    """Test updating analysis jobs when given squeue results."""
    # GIVEN an analysis and a squeue stream
    analysis: Analysis = analysis_store.get_query(table=Analysis).first()
    assert not analysis.jobs

    # WHEN updating the analysis
    analysis_store.update_analysis_from_slurm_run_status(analysis_id=analysis.id)
    updated_analysis: Analysis = analysis_store.get_analysis(
        case_name=analysis.family,
        started_at=analysis.started_at,
        status=TrailblazerStatus.RUNNING,
    )

    # THEN it should update the analysis jobs
    assert updated_analysis.jobs


@pytest.mark.parametrize(
    "family, expected_bool",
    [
        ("blazinginsect", True),  # running
        ("nicemice", False),  # completed
        ("lateraligator", False),  # failed
        ("escapedgoat", True),  # pending
    ],
)
def test_is_latest_analysis_ongoing(analysis_store: MockStore, family: str, expected_bool: bool):
    # GIVEN an analysis
    analysis_store.update_ongoing_analyses()
    analysis_objs = analysis_store.analyses(case_id=family).first()
    assert analysis_objs is not None

    # WHEN checking if the family has an ongoing analysis status
    is_ongoing = analysis_store.is_latest_analysis_ongoing(case_id=family)

    # THEN it should return the expected result
    assert is_ongoing is expected_bool


def test_set_analysis_uploaded(analysis_store: MockStore, timestamp_now: datetime):
    """Test setting analysis uploaded at for an analysis."""

    # GIVEN a store with an analysis
    analysis_obj: Analysis = analysis_store.analyses().first()
    uploaded_at: datetime = timestamp_now

    # WHEN setting an analysis uploaded at
    analysis_store.set_analysis_uploaded(case_id=analysis_obj.family, uploaded_at=uploaded_at)

    # THEN the column uploaded_at should be updated
    assert analysis_obj.uploaded_at == uploaded_at


def test_set_analysis_failed(analysis_store: MockStore):
    """Test setting analysis to failed for an analysis."""

    # GIVEN a store with an analysis
    analysis_obj: Analysis = analysis_store.analyses().first()

    # WHEN setting analysis to failed
    analysis_store.set_analysis_status(case_id=analysis_obj.family, status="failed")

    # THEN the column status should be updated with failed.
    assert analysis_obj.status == "failed"


def test_add_comment(analysis_store: MockStore):
    """Test adding comment to an analysis object."""

    # GIVEN a store with an analysis
    analysis_obj: Analysis = analysis_store.analyses().first()
    comment: str = "test comment"

    # WHEN adding a comment
    analysis_store.add_comment(case_id=analysis_obj.family, comment=comment)

    # THEN a comment should have been added
    assert analysis_obj.comment == comment


@pytest.mark.parametrize(
    "family, expected_bool",
    [
        ("blazinginsect", False),  # running
        ("nicemice", False),  # completed
        ("lateraligator", True),  # failed
        ("escapedgoat", False),  # pending
    ],
)
def test_is_latest_analysis_failed(analysis_store: MockStore, family: str, expected_bool: bool):
    # GIVEN an analysis
    analysis_store.update_ongoing_analyses()
    analysis_objs = analysis_store.analyses(case_id=family).first()
    assert analysis_objs is not None

    # WHEN checking if the family has a failed analysis status
    is_failed = analysis_store.is_latest_analysis_failed(case_id=family)

    # THEN it should return the expected result
    assert is_failed is expected_bool


@pytest.mark.parametrize(
    "family, expected_bool",
    [
        ("blazinginsect", False),  # running
        ("nicemice", True),  # completed
        ("lateraligator", False),  # failed
        ("escapedgoat", False),  # pending
    ],
)
def test_is_latest_analysis_completed(analysis_store: MockStore, family: str, expected_bool: bool):
    # GIVEN an analysis
    analysis_store.update_ongoing_analyses()
    analysis_objs = analysis_store.analyses(case_id=family).first()
    assert analysis_objs is not None

    # WHEN checking if the family has a failed analysis status
    is_failed = analysis_store.is_latest_analysis_completed(case_id=family)

    # THEN it should return the expected result
    assert is_failed is expected_bool


@pytest.mark.parametrize(
    "family, expected_status",
    [
        ("blazinginsect", "running"),  # running
        ("nicemice", "completed"),  # completed
        ("lateraligator", "failed"),  # failed
        ("escapedgoat", "pending"),  # pending
    ],
)
def test_get_latest_analysis_status(analysis_store: MockStore, family: str, expected_status: str):
    # GIVEN an analysis
    analysis_store.update_ongoing_analyses()
    analysis_objs = analysis_store.analyses(case_id=family).first()
    assert analysis_objs is not None

    # WHEN checking if the family has an analysis status
    status = analysis_store.get_latest_analysis_status(case_id=family)

    # THEN it should return the expected result
    assert status is expected_status


@pytest.mark.parametrize(
    "case_id, status",
    [
        ("blazinginsect", "running"),
        ("crackpanda", "failed"),
        ("daringpidgeon", "error"),
        ("emptydinosaur", "error"),
        ("escapedgoat", "pending"),
        ("fancymole", "completed"),
        ("happycow", "pending"),
        ("lateraligator", "failed"),
        ("liberatedunicorn", "error"),
        ("nicemice", "completed"),
        ("rarekitten", "canceled"),
        ("trueferret", "running"),
    ],
)
def test_update(analysis_store: MockStore, case_id, status):
    # GIVEN an analysis
    analysis_obj = analysis_store.get_latest_analysis(case_id)

    # WHEN database is updated once
    analysis_store.update_run_status(analysis_obj.id)

    # THEN analysis status is changed to what is expected
    assert analysis_obj.status == status

    # WHEN database is updated a second time
    analysis_store.update_run_status(analysis_obj.id)

    # THEN the status is still what is expected, and no database errors were raised
    assert analysis_obj.status == status


def test_mark_analyses_deleted(analysis_store: MockStore):
    # GIVEN case_id for a case that is not deleted
    case_id = "liberatedunicorn"
    analysis_obj = analysis_store.get_latest_analysis(case_id)
    assert not analysis_obj.is_deleted

    # WHEN running command
    analysis_store.mark_analyses_deleted(case_id=case_id)
    analysis_obj = analysis_store.get_latest_analysis(case_id)

    # THEN analysis is marked deleted
    assert analysis_obj.is_deleted


def test_update_tower_jobs(analysis_store: MockStore, tower_jobs: List[dict], case_id: str):
    """Assess that jobs are successfully updated when using NF Tower."""

    # GIVEN an analysis without failed jobs
    analysis: Analysis = analysis_store.get_latest_analysis(case_id=case_id)
    assert not analysis.jobs

    # WHEN analysis jobs are deleted
    analysis_store.delete_analysis_jobs(analysis=analysis)

    # THEN analysis object should have no failed jobs
    assert not analysis.jobs

    # WHEN jobs are updated
    analysis_store.update_analysis_jobs(analysis=analysis, jobs=tower_jobs[:2])

    # THEN failed jobs should be updated
    assert len(analysis.jobs) == 2


@pytest.mark.parametrize(
    "case_id, status, progress",
    [
        (CaseIDs.RUNNING, TrailblazerStatus.RUNNING, 0.15),
        (CaseIDs.PENDING, TrailblazerStatus.PENDING, 0),
        (CaseIDs.COMPLETED, TrailblazerStatus.QC, 1),
    ],
)
def test_update_tower_run_status(
    analysis_store: MockStore, case_id: str, status: str, progress: int
):
    """Assess that an analysis status is successfully updated when using NF Tower."""

    # GIVEN an analysis with pending status
    analysis: Analysis = analysis_store.get_latest_analysis(case_id=case_id)
    assert analysis.status == TrailblazerStatus.PENDING

    # WHEN database is updated once
    analysis_store.update_run_status(analysis_id=analysis.id)

    # THEN analysis status is changed to what is expected
    assert analysis.status == status
    assert analysis.progress == progress

    # WHEN database is updated a second time
    analysis_store.update_run_status(analysis_id=analysis.id)

    # THEN the status is still as before, and no database errors were raised
    assert analysis.status == status
    assert analysis.progress == progress



from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, List, Generator

import pytest

from tests.apps.tower.conftest import CaseIDs, TowerTaskResponseFile
from tests.mocks.store_mock import MockStore
from trailblazer.apps.tower.models import TowerTask
from trailblazer.constants import TOWER_TIMESTAMP_FORMAT, TrailblazerStatus, FileFormat
from trailblazer.io.controller import ReadFile
from tests.store.utils.store_helper import StoreHelpers


@pytest.fixture(scope="session", name="username")
def fixture_username() -> str:
    """Return a username."""
    return "Paul Anderson"


@pytest.fixture(scope="session", name="archived_username")
def fixture_archived_username() -> str:
    """Return an archived username."""
    return "Archived User"


@pytest.fixture(scope="session", name="archived_user_email")
def fixture_archived_user_email() -> str:
    """Return an archived user email."""
    return "archived.user@magnolia.com"


@pytest.fixture(scope="session", name="user_email")
def fixture_user_email() -> str:
    """Return an user email."""
    return "paul.anderson@magnolia.com"


@pytest.fixture(scope="session", name="fixtures_dir")
def fixture_fixtures_dir() -> Path:
    """Return the path to the fixtures' dir."""
    return Path("tests", "fixtures")


@pytest.fixture(scope="session", name="analysis_data_path")
def fixture_analysis_data_path(fixtures_dir: Path) -> Path:
    """Return the path to an analysis data file."""
    return Path(fixtures_dir, "analysis-data.yaml")


@pytest.fixture(name="analysis_data", scope="function")
def fixture_analysis_data(analysis_data_path: Path) -> Dict[str, list]:
    """Return content of the analysis data file."""
    return ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=analysis_data_path)


@pytest.fixture(scope="session", name="squeue_stream_jobs")
def fixture_squeue_stream_jobs() -> str:
    """Return a squeue output stream."""
    return """JOBID,NAME,STATE,TIME_LIMIT,TIME,START_TIME
690994,gatk_genotypegvcfs2,COMPLETED,10:00:00,1:01:52,2020-10-22T11:43:33
702463,bwa_mem_ACC3218A1,COMPLETED,1-06:00:00,1-1:28:12,2020-10-27T23:06:34
690992,gatk_genotypegvcfs3,COMPLETED,10:00:00,5:54,2020-10-22T11:43:02
690988,gatk_genotypegvcfs4,RUNNING,10:00:00,0:19,N/A
690989,gatk_genotypegvcfs5,PENDING,10:00:00,0:00,N/A"""


@pytest.fixture(name="trailblazer_tmp_dir")
def fixture_trailblazer_tmp_dir(tmpdir_factory) -> Path:
    """Return a temporary directory for Trailblazer testing."""
    return tmpdir_factory.mktemp("trailblazer_tmp")


@pytest.fixture(name="trailblazer_context")
def fixture_trailblazer_context(analysis_store: MockStore) -> Dict[str, MockStore]:
    """Trailblazer context to be used in CLI."""
    return {"trailblazer": analysis_store}


@pytest.fixture(name="store")
def fixture_store() -> Generator[MockStore, None, None]:
    """Empty Trailblazer database."""
    _store = MockStore(uri="sqlite://")
    _store.setup()
    yield _store
    _store.drop_all()


@pytest.fixture(scope="function", name="info_store")
def fixture_info_store(store: MockStore) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated info table."""
    StoreHelpers.add_info(store=store)
    yield store


@pytest.fixture(scope="function", name="job_store")
def fixture_job_store(
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated job table."""
    statuses: List[str] = [TrailblazerStatus.COMPLETED, TrailblazerStatus.FAILED]
    for index, status in enumerate(statuses):
        StoreHelpers.add_job(
            analysis_id=index, name=str(index), slurm_id=index, status=status, store=store
        )
    yield store


@pytest.fixture(scope="function", name="user_store")
def fixture_user_store(
    archived_user_email: str,
    archived_username: str,
    user_email: str,
    username: str,
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated user table."""
    StoreHelpers.add_user(email=user_email, name=username, store=store)
    StoreHelpers.add_user(
        email=archived_user_email, name=archived_username, is_archived=True, store=store
    )
    yield store


@pytest.fixture(name="raw_analyses", scope="function")
def fixture_raw_analyses(analysis_data: Dict[str, List[Dict]]) -> List[dict]:
    """Return raw analyses data."""
    analyses: List[dict] = []
    for analysis in analysis_data["analyses"]:
        analysis["case_id"] = analysis["family"]
        analyses.append(analysis)
    return analyses


@pytest.fixture(name="analysis_store")
def fixture_analysis_store(
    analysis_data: Dict[str, list],
    archived_user_email: str,
    archived_username: str,
    raw_analyses: List[dict],
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A sample Trailblazer database populated with pending analyses."""
    StoreHelpers.add_user(
        email=archived_user_email, name=archived_username, is_archived=True, store=store
    )
    for user_data in analysis_data["users"]:
        store.add_user(name=user_data["name"], email=user_data["email"])
    for raw_analysis in raw_analyses:
        raw_analysis["user"] = store.get_user(email=raw_analysis["user"])
        store.add(store.Analysis(**raw_analysis))
    store.commit()
    yield store


@pytest.fixture(name="timestamp_now", scope="session")
def fixture_timestamp_now() -> datetime:
    """Return a time stamp of today's date in date time format."""
    return datetime.now()


@pytest.fixture(name="timestamp_yesterday", scope="session")
def fixture_timestamp_yesterday(timestamp_now: datetime) -> datetime:
    """Return a time stamp of yesterday's date in date time format."""
    return timestamp_now - timedelta(days=1)


@pytest.fixture(name="tower_jobs")
def fixture_tower_jobs(analysis_id, started_at, slurm_id, tower_task_name) -> List[dict]:
    """Return a list of Tower Jobs."""
    return [
        dict(
            analysis_id=analysis_id,
            slurm_id=slurm_id,
            name=tower_task_name,
            started_at=started_at,
            elapsed=63,
            status=TrailblazerStatus.COMPLETED,
        ),
        dict(
            analysis_id=analysis_id,
            slurm_id="4611829",
            name="NFCORE_RNAFUSION:RNAFUSION:PIZZLY_WORKFLOW:KALLISTO_QUANT",
            started_at=None,
            elapsed=0,
            status=TrailblazerStatus.PENDING,
        ),
        dict(
            analysis_id=analysis_id,
            slurm_id="4611828",
            name="NFCORE_RNAFUSION:RNAFUSION:FASTQC",
            started_at=None,
            elapsed=0,
            status=TrailblazerStatus.PENDING.value,
        ),
    ]


@pytest.fixture(name="analysis_id")
def fixture_analysis_id() -> int:
    """Return a mock ID of the analysis in the Trailblazer database."""
    return 1


@pytest.fixture(name="started_at", scope="session")
def fixture_started_at() -> datetime:
    """Returns a started at date."""
    return datetime.strptime("2023-04-04T08:11:27Z", TOWER_TIMESTAMP_FORMAT)


@pytest.fixture(name="slurm_id", scope="session")
def fixture_slurm_id() -> str:
    """Returns a slurm id."""
    return "4611827"


@pytest.fixture(name="tower_task_name", scope="session")
def fixture_tower_task_name() -> str:
    """Returns a NF Tower task name."""
    return "NFCORE_RNAFUSION:RNAFUSION:INPUT_CHECK:SAMPLESHEET_CHECK"


@pytest.fixture(name="case_id", scope="session")
def fixture_case_id() -> str:
    """Return a case ID."""
    return CaseIDs.RUNNING


@pytest.fixture(name="tower_task", scope="session")
def fixture_tower_task() -> TowerTask:
    """Return a Tower task."""
    tower_task_running_content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=TowerTaskResponseFile.RUNNING
    )
    return TowerTask(**tower_task_running_content["tasks"][0]["task"])



from functools import partial

import pytest
from click.testing import CliRunner

from trailblazer.cli.core import base


@pytest.fixture(name="process_exit_success", scope="session")
def fixture_process_exit_success() -> int:
    """Return the exit code for a successful process."""
    return 0


@pytest.fixture(name="cli_runner", scope="session")
def fixture_cli_runner() -> CliRunner:
    """Return a CliRunner fixture."""
    return CliRunner()


@pytest.fixture(name="invoke_cli", scope="session")
def fixture_invoke_cli(cli_runner: CliRunner) -> partial:
    """Invokes CLI base with partial functionality."""
    return partial(cli_runner.invoke, base)



import logging
from datetime import datetime, timedelta
from typing import List

from trailblazer.constants import (
    HOURS_IN_DAY,
    MINUTES_PER_HOUR,
    SECONDS_PER_MINUTE,
    TOWER_TIMESTAMP_FORMAT,
    TOWER_TIMESTAMP_FORMAT_EXTENDED,
)

LOG = logging.getLogger(__name__)


def convert_days_to_minutes(days_nr: int) -> int:
    """Converts number of days to minutes."""
    return days_nr * HOURS_IN_DAY * MINUTES_PER_HOUR


def convert_timestamp_to_minutes(timestamp: datetime) -> int:
    """Converts timestamp to minutes."""
    zero_seconds: datetime = datetime.strptime("0:0:0", "%H:%M:%S")
    return int((timestamp - zero_seconds).total_seconds() / SECONDS_PER_MINUTE)


def get_datetime_from_timestamp(timestamp: str, datetime_formats: List[str]) -> datetime:
    """Converts a timestamp into a datatime object."""
    for datetime_format in datetime_formats:
        try:
            return datetime.strptime(timestamp, datetime_format)
        except ValueError as error:
            error_message = f"Error converting timestamp: {error}"
            continue
    LOG.error(error_message)


def tower_datetime_converter(timestamp: str) -> datetime:
    """Converts a NF Tower timestamp into a datetime object."""
    allowed_formats = [TOWER_TIMESTAMP_FORMAT, TOWER_TIMESTAMP_FORMAT_EXTENDED]
    return get_datetime_from_timestamp(timestamp, allowed_formats)


def get_date_number_of_days_ago(number_of_days_ago: int) -> datetime:
    """Return the date that was number of 'days_ago'."""
    return datetime.now() - timedelta(days=number_of_days_ago)



from typing import List

from tests.mocks.store_mock import MockStore
from trailblazer.store.filters.user_filters import UserFilter, apply_user_filter
from trailblazer.store.models import Analysis, User


def test_add_pending_analysis(raw_analyses: List[dict], store: MockStore, user_email: str):
    """Test adding a new analysis to the database."""
    # GIVEN an empty database
    assert not store.get_query(table=Analysis).first()

    # GIVEN an unprocessed raw analysis
    analysis: dict = raw_analyses[0]

    # WHEN adding a new analysis
    new_analysis: Analysis = store.add_pending_analysis(
        case_id=analysis.get("family"),
        config_path=analysis.get("config_path"),
        email=user_email,
        out_dir=analysis.get("out_dir"),
        priority=analysis.get("priority"),
        type=analysis.get("type"),
    )

    # THEN it should be stored in the database
    stored_analysis: Analysis = store.get_analysis(
        case_name=analysis.get("family"),
        started_at=new_analysis.started_at,
        status=analysis.get("status"),
    )
    assert new_analysis and stored_analysis
    assert stored_analysis == new_analysis


def test_add_user(store: MockStore, user_email: str, username: str):
    """Test adding a user to the database."""
    # GIVEN an empty database
    assert not store.get_query(table=User).first()

    # WHEN adding a new user
    new_user: User = store.add_user(name=username, email=user_email)

    # THEN it should be stored in the database
    user: User = apply_user_filter(
        filter_functions=[UserFilter.FILTER_BY_EMAIL],
        users=store.get_query(table=User),
        email=user_email,
    ).first()
    assert new_user
    assert user == new_user



from pathlib import Path
from typing import List

import pytest

from tests.apps.tower.conftest import TowerResponseFile, TowerTaskResponseFile
from tests.mocks.tower_mock import MockTowerAPI
from trailblazer.apps.tower.models import TowerTask
from trailblazer.constants import TrailblazerStatus


@pytest.mark.parametrize(
    "response_file, expected_status",
    [
        (TowerResponseFile.PENDING, TrailblazerStatus.PENDING),
        (TowerResponseFile.RUNNING, TrailblazerStatus.RUNNING),
        (TowerResponseFile.COMPLETED, TrailblazerStatus.QC),
    ],
)
def test_tower_api_status(tower_id: str, response_file: Path, expected_status: str) -> None:
    """Assess that TowerAPI returns the correct status given a response."""

    # GIVEN an tower_api with a mock query response
    tower_api = MockTowerAPI(workflow_id=tower_id)
    tower_api.mock_query(response_file=response_file)

    # THEN a trailblazer status is returned
    assert tower_api.status == expected_status


@pytest.mark.parametrize(
    "response_file, expected_bool",
    [
        (TowerResponseFile.PENDING, True),
        (TowerResponseFile.RUNNING, False),
        (TowerResponseFile.COMPLETED, False),
    ],
)
def test_tower_api_is_pending(tower_id: str, response_file: Path, expected_bool: bool) -> None:
    """Assess that TowerAPI returns the state for is_pending given a response."""

    # GIVEN an tower_api with a mock query response
    tower_api = MockTowerAPI(workflow_id=tower_id)
    tower_api.mock_query(response_file=response_file)

    # THEN is_pending should be returned
    assert tower_api.is_pending == expected_bool


@pytest.mark.parametrize(
    "response_file, expected_nr_total_jobs",
    [
        (TowerResponseFile.PENDING, 0),
        (TowerResponseFile.RUNNING, 13),
        (TowerResponseFile.COMPLETED, 13),
    ],
)
def test_tower_api_total_jobs(
    tower_id: str, response_file: Path, expected_nr_total_jobs: int
) -> None:
    """Assess that TowerAPI correctly returns the total number of potential jobs given a response."""

    # GIVEN an tower_api with a mock query response
    tower_api = MockTowerAPI(workflow_id=tower_id)
    tower_api.mock_query(response_file=response_file)

    # THEN the total number of potential jobs is returned
    assert tower_api.total_jobs == expected_nr_total_jobs


@pytest.mark.parametrize(
    "response_file, expected_succeeded_jobs",
    [
        (TowerResponseFile.PENDING, 0),
        (TowerResponseFile.RUNNING, 2),
        (TowerResponseFile.COMPLETED, 6),
    ],
)
def test_tower_api_succeeded_jobs(
    tower_id: str, response_file: Path, expected_succeeded_jobs: int
) -> None:
    """Assess that TowerAPI correctly returns the number of succeeded jobs given a response."""

    # GIVEN an tower_api with a mock query response
    tower_api = MockTowerAPI(workflow_id=tower_id)
    tower_api.mock_query(response_file=response_file)

    # THEN the total number of succeeded jobs is returned
    assert tower_api.succeeded_jobs == expected_succeeded_jobs


def test_tower_api_tasks(
    tower_id: str,
    analysis_id: int,
    tower_jobs: List[dict],
    caplog,
) -> None:
    """Assess that TowerAPI returns a list of tasks given a response."""

    # GIVEN an tower_api with a mock query response
    tower_api = MockTowerAPI(workflow_id=tower_id)
    tower_api.mock_tasks_query(response_file=TowerTaskResponseFile.RUNNING)

    # WHEN asking for jobs
    jobs = tower_api.get_jobs(analysis_id=analysis_id)

    # THEN a list of jobs should be returned
    for job_nr in range(1, len(jobs)):
        assert dict(jobs[job_nr]) == dict(tower_jobs[job_nr])

    # THEN no error should be logged when converting timestamps
    assert "Error converting timestamp" not in caplog.text


@pytest.mark.parametrize(
    "response_file, expected_progress",
    [
        (TowerResponseFile.PENDING, 0),
        (TowerResponseFile.RUNNING, 0.15),
        (TowerResponseFile.COMPLETED, 1),
    ],
)
def test_tower_api_progress(tower_id: str, response_file: Path, expected_progress: float) -> None:
    """Assess that TowerAPI returns the progress percentage given a response."""

    # GIVEN an tower_api with a mock query response
    tower_api = MockTowerAPI(workflow_id=tower_id)
    tower_api.mock_query(response_file=response_file)

    # THEN progress should be returned
    assert tower_api.progress == expected_progress


def test_tower_api_tasks_empty(
    tower_id: str,
    analysis_id: int,
    tower_task_response_pending: Path,
) -> None:
    """Assess that TowerAPI returns an empty list of tasks given a response for a pending case."""

    # GIVEN an tower_api with a mock query response for a pending case
    tower_api = MockTowerAPI(workflow_id=tower_id)
    tower_api.mock_tasks_query(response_file=tower_task_response_pending)

    # WHEN asking for jobs
    jobs = tower_api.get_jobs(analysis_id=analysis_id)

    # THEN an empty list should be returned
    assert jobs == []


def test_tower_task_properties(
    tower_task: TowerTask,
    created_at,
    started_at,
    last_updated,
    tower_task_duration,
    slurm_id,
    tower_task_name,
) -> None:
    """Assess that TowerTask returns the right properties."""

    # GIVEN a tower task

    # THEN properties should be returned
    assert tower_task.process == tower_task_name
    assert tower_task.nativeId == slurm_id
    assert tower_task.status == TrailblazerStatus.COMPLETED
    assert tower_task.dateCreated == created_at
    assert tower_task.lastUpdated == last_updated
    assert tower_task.start == started_at
    assert tower_task.duration == tower_task_duration
    assert tower_task.is_complete is True



from datetime import datetime
from enum import Enum
from typing import Callable, List, Optional

from sqlalchemy.orm import Query

from trailblazer.constants import TrailblazerStatus
from trailblazer.store.models import Analysis


def filter_analyses_by_id(analyses: Query, analysis_id: int, **kwargs) -> Query:
    """Filter analyses by database entry id."""
    return analyses.filter(Analysis.id == analysis_id)


def filter_analyses_by_case_name(analyses: Query, case_name: str, **kwargs) -> Query:
    """Filter analyses by case name."""
    return analyses.filter(Analysis.family == case_name)


def filter_analyses_by_started_at(analyses: Query, started_at: datetime, **kwargs) -> Query:
    """Filter analyses by when it started."""
    return analyses.filter(Analysis.started_at == started_at)


def filter_analyses_by_status(analyses: Query, status: TrailblazerStatus, **kwargs) -> Query:
    """Filter analyses by status."""
    return analyses.filter(Analysis.status == status)


class AnalysisFilter(Enum):
    """Define Analysis filter functions."""

    FILTER_BY_ID: Callable = filter_analyses_by_id
    FILTER_BY_CASE_NAME: Callable = filter_analyses_by_case_name
    FILTER_BY_STARTED_AT: Callable = filter_analyses_by_started_at
    FILTER_BY_STATUS: Callable = filter_analyses_by_status


def apply_analysis_filter(
    analyses: Query,
    filter_functions: List[Callable],
    analysis_id: Optional[int] = None,
    case_name: Optional[str] = None,
    started_at: Optional[datetime] = None,
    status: Optional[str] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        analyses: Query = function(
            analyses=analyses,
            analysis_id=analysis_id,
            case_name=case_name,
            started_at=started_at,
            status=status,
        )
    return analyses



from sqlalchemy.orm import Query

from tests.mocks.store_mock import MockStore
from trailblazer.store.filters.user_filters import (
    filter_users_by_email,
    filter_users_by_contains_email,
    filter_users_by_contains_name,
    filter_users_by_is_not_archived,
)
from trailblazer.store.models import User
from tests.store.utils.store_helper import StoreHelpers


def test_filter_users_by_email(user_store: MockStore, user_email: str):
    """Test return user by email."""
    # GIVEN a store containing a user

    # WHEN retrieving a user by email
    users: Query = filter_users_by_email(users=user_store.get_query(table=User), email=user_email)

    # ASSERT that the users is a query
    assert isinstance(users, Query)

    # THEN the users attribute email should match the original
    assert users[0].email == user_email


def test_filter_users_by_contains_email(user_store: MockStore, user_email: str):
    """Test return user by email when using an incomplete email."""
    # GIVEN a store containing a user

    # WHEN retrieving a user by email
    users: Query = filter_users_by_contains_email(
        users=user_store.get_query(table=User), email=user_email[4]
    )

    # ASSERT that the users is a query
    assert isinstance(users, Query)

    # THEN the users attribute email should match the original
    assert users[0].email == user_email


def test_filter_users_by_contains_name(user_store: MockStore, username: str):
    """Test return user which contains name."""
    # GIVEN a store containing a user

    # WHEN retrieving a user by name
    users: Query = filter_users_by_contains_name(
        users=user_store.get_query(table=User), name=username
    )

    # THEN the users is a query
    assert isinstance(users, Query)

    # THEN the users attribute name should match the original
    assert users[0].name == username


def test_filter_users_by_not_archived(user_store: MockStore, username: str):
    """Test return user which contains name and is not archived."""
    # GIVEN a store containing a not archived user and an archived user
    StoreHelpers.add_user(
        name="New User", email="new.user@magnolia.com", store=user_store, is_archived=True
    )

    # WHEN retrieving a not archived user by name
    users: Query = filter_users_by_is_not_archived(
        users=user_store.get_query(table=User), name=username
    )

    # THEN the users is a query
    assert isinstance(users, Query)

    # THEN only one user should be returned
    assert users.count() == 1

    # THEN the users attribute name should match the original
    assert users[0].name == username



from enum import Enum
from typing import Dict, List, Tuple

ONE_MONTH_IN_DAYS: int = 31
HOURS_IN_DAY: int = 24
MINUTES_PER_HOUR: int = 60
SECONDS_PER_MINUTE: int = 60
PRIORITY_OPTIONS: Tuple = ("low", "normal", "high", "express", "maintenance")
TRAILBLAZER_TIME_STAMP: str = "%Y-%m-%d"
TOWER_TIMESTAMP_FORMAT: str = "%Y-%m-%dT%H:%M:%SZ"
TOWER_TIMESTAMP_FORMAT_EXTENDED: str = "%Y-%m-%dT%H:%M:%S.%fZ"
TYPES: Tuple = ("other", "rna", "tgs", "wes", "wgs", "wts")


class FileFormat(str, Enum):
    CSV: str = "csv"
    JSON: str = "json"
    YAML: str = "yaml"


class FileExtension(str, Enum):
    CSV: str = ".csv"
    JSON: str = ".json"
    YAML: str = ".yaml"


class WorkflowManager(Enum):
    """Supported task managers."""

    SLURM: str = "slurm"
    TOWER: str = "nf_tower"

    @classmethod
    def list(cls) -> List:
        return [task.value for task in cls]


class Pipeline(str, Enum):
    """Analysis pipeline names."""

    BALSAMIC: str = "BALSAMIC"
    MIP_DNA: str = "MIP-DNA"
    MIP_RNA: str = "MIP-RNA"
    SARS_COV_2: str = "SARS-COV-2"


class SlurmSqueueHeader(str, Enum):
    """SLURM squeue output headers."""

    JOBID: str = "JOBID"
    NAME: str = "NAME"
    START_TIME: str = "START_TIME"
    STATE: str = "STATE"
    TIME: str = "TIME"
    TIME_LIMIT: str = "TIME_LIMIT"


class SlurmJobStatus(str, Enum):
    """SLURM allowed status."""

    CANCELLED: str = "cancelled"
    COMPLETED: str = "completed"
    COMPLETING: str = "completing"
    FAILED: str = "failed"
    PENDING: str = "pending"
    RUNNING: str = "running"
    TIME_OUT: str = "timeout"

    @classmethod
    def statuses(cls) -> Tuple:
        return tuple(status.value for status in cls)

    @classmethod
    def ongoing_statuses(cls) -> Tuple:
        return cls.PENDING.value, cls.RUNNING.value, cls.COMPLETING.value


class TrailblazerStatus(str, Enum):
    """Trailblazer allowed status."""

    CANCELLED: str = "canceled"
    COMPLETED: str = "completed"
    COMPLETING: str = "completing"
    ERROR: str = "error"
    FAILED: str = "failed"
    PENDING: str = "pending"
    QC: str = "qc"
    RUNNING: str = "running"

    @classmethod
    def statuses(cls) -> Tuple:
        return tuple(status.value for status in cls)

    @classmethod
    def ongoing_statuses(cls) -> Tuple:
        return cls.PENDING.value, cls.RUNNING.value, cls.COMPLETING.value, cls.ERROR.value


class TrailblazerStatusColor(str, Enum):
    """Trailblazer status colors."""

    COMPLETED: str = "green"
    DEFAULT: str = "white"
    FAILED: str = "red"
    PENDING: str = "yellow"
    RUNNING: str = "blue"


TOWER_STATUS: Dict[str, str] = {
    "ABORTED": TrailblazerStatus.FAILED,
    "CACHED": TrailblazerStatus.COMPLETED,
    "CANCELLED": TrailblazerStatus.FAILED,
    "COMPLETED": TrailblazerStatus.COMPLETED,
    "FAILED": TrailblazerStatus.FAILED,
    "NEW": TrailblazerStatus.PENDING,
    "RUNNING": TrailblazerStatus.RUNNING,
    "SUBMITTED": TrailblazerStatus.PENDING,
    "SUCCEEDED": TrailblazerStatus.COMPLETED,
    "UNKNOWN": TrailblazerStatus.FAILED,
}


TOWER_PROCESS_STATUS: Dict[str, str] = {
    "submitted": TrailblazerStatus.PENDING,
    "pending": TrailblazerStatus.PENDING,
    "running": TrailblazerStatus.RUNNING,
    "cached": TrailblazerStatus.COMPLETED,
    "succeeded": TrailblazerStatus.COMPLETED,
    "failed": TrailblazerStatus.FAILED,
}



import datetime as dt
from pathlib import Path

import pytest

from trailblazer.constants import TOWER_TIMESTAMP_FORMAT, TOWER_TIMESTAMP_FORMAT_EXTENDED

TOWER_RESPONSE_DIR: Path = Path("tests", "fixtures", "tower")
TOWER_ID: str = "1m759EPcbjuK7n"


class TowerResponseFile:
    PENDING: Path = Path(TOWER_RESPONSE_DIR, "tower_workflow_pending.json")
    RUNNING: Path = Path(TOWER_RESPONSE_DIR, "tower_workflow_running.json")
    COMPLETED: Path = Path(TOWER_RESPONSE_DIR, "tower_workflow_completed.json")
    EMPTY: Path = Path(TOWER_RESPONSE_DIR, "tower_workflow_empty.json")


class TowerTaskResponseFile:
    PENDING: Path = Path(TOWER_RESPONSE_DIR, "tower_tasks_pending.json")
    RUNNING: Path = Path(TOWER_RESPONSE_DIR, "tower_tasks_running.json")
    COMPLETED: Path = Path(TOWER_RESPONSE_DIR, "tower_tasks_completed.json")
    EMPTY: Path = Path(TOWER_RESPONSE_DIR, "tower_tasks_empty.json")


class CaseIDs:
    PENDING: str = "cuddlyhen_pending"
    RUNNING: str = "cuddlyhen"
    COMPLETED: str = "cuddlyhen_completed"


@pytest.fixture(name="tower_id", scope="session")
def fixture_tower_id() -> str:
    """Return a NF Tower id."""
    return TOWER_ID


@pytest.fixture(name="tower_config_file", scope="session")
def fixture_tower_config_file() -> str:
    """Return the path of a config yaml file with a NF Tower id."""
    return Path("tests", "fixtures", "case", "cuddlyhen_tower_id.yaml").as_posix()


@pytest.fixture(name="tower_task_response_pending", scope="session")
def fixture_tower_task_response_pending() -> Path:
    """Return an NF Tower task response for a pending case."""
    return TowerTaskResponseFile.PENDING


@pytest.fixture(name="created_at")
def fixture_created_at() -> dt.datetime:
    """Returns a created at date."""
    return dt.datetime.strptime("2023-04-04T08:11:24Z", TOWER_TIMESTAMP_FORMAT)


@pytest.fixture(name="last_updated")
def fixture_last_updated() -> dt.datetime:
    """Returns a last updated date."""
    return dt.datetime.strptime("2023-06-20T08:01:57.661819Z", TOWER_TIMESTAMP_FORMAT_EXTENDED)


@pytest.fixture(name="tower_task_duration")
def fixture_tower_task_duration() -> int:
    """Returns the duration of a NF Tower task."""
    return 4



"""Module for reading and writing comma separated values (CSV) formatted files."""
import csv
from typing import List


def read_csv_stream(stream: bytes, read_to_dict: bool = False) -> List[List[str]]:
    """Read CSV formatted stream."""
    csv_reader = (
        csv.DictReader(stream.splitlines()) if read_to_dict else csv.reader(stream.splitlines())
    )
    return list(csv_reader)



from trailblazer.store.base import BaseHandler_2
from trailblazer.store.models import Analysis


class DeleteHandler(BaseHandler_2):
    """Class for deleting items in the database."""

    def delete_analysis_jobs(self, analysis: Analysis) -> None:
        """Delete all jobs linked to the given analysis."""
        for job in analysis.jobs:
            job.delete()
        self.commit()



from typing import Dict, Tuple

from trailblazer.constants import TrailblazerStatus, TrailblazerStatusColor
from trailblazer.store.models import Analysis


def _get_ls_analysis_message(analysis: Analysis) -> Tuple[str, str]:
    """Return the log message for the analysis."""
    message: str = (
        f"{analysis.id} | {analysis.family} {analysis.started_at.date()} "
        f"[{analysis.type.upper()}/{analysis.status.upper()}]"
    )
    message_map: Dict[str, tuple] = {
        TrailblazerStatus.PENDING: (
            f"{analysis.id} | {analysis.family} [{analysis.status.upper()}]",
            TrailblazerStatusColor.PENDING,
        ),
        TrailblazerStatus.RUNNING: (
            f"{message} - {analysis.progress * 100}/100",
            TrailblazerStatusColor.RUNNING,
        ),
        TrailblazerStatus.COMPLETED: (
            f"{message} - {analysis.completed_at}",
            TrailblazerStatusColor.COMPLETED,
        ),
        TrailblazerStatus.FAILED: (message, TrailblazerStatusColor.FAILED),
    }
    return message_map.get(analysis.status, (message, TrailblazerStatusColor.DEFAULT))



from datetime import datetime
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, field_validator

from trailblazer.constants import TOWER_PROCESS_STATUS, TOWER_STATUS, TrailblazerStatus
from trailblazer.utils.datetime import tower_datetime_converter

SCALE_TO_MILLISEC: int = 1000


class TowerTask(BaseModel):
    """NF Tower task model."""

    process: str
    name: str
    status: str
    nativeId: str
    dateCreated: Optional[Union[str, datetime]] = None
    lastUpdated: Optional[Union[str, datetime]] = None
    start: Optional[Union[str, datetime]] = None
    duration: Optional[int] = None
    hash: Optional[str] = None
    tag: Optional[str] = None
    submit: Optional[str] = None
    complete: Optional[str] = None
    module: List[str]
    container: Optional[str] = None
    attempt: Optional[int] = None
    script: Optional[str] = None
    scratch: Optional[str] = None
    workdir: Optional[str] = None
    queue: Optional[str] = None
    cpus: Optional[int] = None
    memory: Optional[int] = None
    disk: Optional[int] = None
    time: Optional[int] = None
    env: Optional[str] = None
    executor: Optional[str] = None
    machineType: Optional[str] = None
    cloudZone: Optional[str] = None
    priceModel: Optional[str] = None
    cost: Optional[float] = None
    errorAction: Optional[str] = None
    exitStatus: Optional[int] = None
    realtime: Optional[int] = None
    pcpu: Optional[float] = None
    pmem: Optional[float] = None
    rss: Optional[int] = None
    vmem: Optional[int] = None
    peakRss: Optional[int] = None
    peakVmem: Optional[int] = None
    rchar: Optional[int] = None
    wchar: Optional[int] = None
    syscr: Optional[int] = None
    syscw: Optional[int] = None
    readBytes: Optional[int] = None
    writeBytes: Optional[int] = None
    volCtxt: Optional[int] = None
    invCtxt: Optional[int] = None
    exit: Optional[str] = None
    id: Optional[int] = None
    taskId: Optional[int] = None
    model_config = ConfigDict(validate_default=True)

    @field_validator("duration")
    @classmethod
    def set_duration(cls, raw_duration: Optional[int]) -> int:
        """Convert milliseconds to seconds or return 0 if empty."""
        return round(raw_duration / SCALE_TO_MILLISEC) if raw_duration else 0

    @field_validator("status")
    @classmethod
    def set_status(cls, raw_status) -> str:
        return TOWER_STATUS.get(raw_status)

    @field_validator("start", "dateCreated", "lastUpdated")
    @classmethod
    def set_datetime(cls, raw_time) -> Optional[Union[str, datetime]]:
        if type(raw_time) is str:
            return tower_datetime_converter(timestamp=raw_time)
        elif type(raw_time) is datetime:
            return raw_time
        else:
            return None

    @property
    def is_complete(cls) -> bool:
        """Returns if the process succeded."""
        return cls.status == TrailblazerStatus.COMPLETED


class TowerProcess(BaseModel):
    """NF Tower task model."""

    process: str
    dateCreated: Optional[Union[str, datetime]] = None
    lastUpdated: Optional[Union[str, datetime]] = None
    pending: Optional[int] = None
    submitted: Optional[int] = None
    running: Optional[int] = None
    succeeded: Optional[int] = None
    failed: Optional[int] = None
    cached: Optional[int] = None
    memoryEfficiency: Optional[float] = None
    cpuEfficiency: Optional[float] = None
    loadCpus: Optional[int] = None
    loadMemory: Optional[int] = None
    peakCpus: Optional[int] = None
    peakTasks: Optional[int] = None
    peakMemory: Optional[int] = None
    model_config = ConfigDict(validate_default=True)

    @field_validator("dateCreated", "lastUpdated")
    @classmethod
    def set_datetime(cls, raw_time) -> Optional[Union[str, datetime]]:
        if type(raw_time) is str:
            return tower_datetime_converter(timestamp=raw_time)
        elif type(raw_time) is datetime:
            return raw_time
        else:
            return None

    @property
    def is_complete(cls) -> bool:
        """Returns if the process succeded."""
        return cls.status == TrailblazerStatus.COMPLETED

    @property
    def status(cls) -> str:
        for status_flag in TOWER_PROCESS_STATUS.keys():
            if cls.dict().get(status_flag, 0):
                return TOWER_PROCESS_STATUS.get(status_flag)
        return TrailblazerStatus.ERROR


class TowerWorkflow(BaseModel):
    """NF Tower workflow model."""

    status: str


class TowerProgress(BaseModel):
    """NF Tower progress model."""

    workflowProgress: Dict
    processesProgress: List


class TowerTaskResponse(BaseModel):
    """NF Tower task response model."""

    tasks: List[Optional[Dict[str, TowerTask]]]
    total: int


class TowerWorkflowResponse(BaseModel):
    """NF Tower task model."""

    workflow: TowerWorkflow
    progress: TowerProgress



"""Model SLURM output."""
from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel, Field, field_validator, model_validator

from trailblazer.constants import SlurmJobStatus, SlurmSqueueHeader
from trailblazer.exc import MissingSqueueOutput
from trailblazer.utils.datetime import (
    convert_days_to_minutes,
    convert_timestamp_to_minutes,
    get_datetime_from_timestamp,
)


class SqueueJob(BaseModel):
    """Model job meta data from squeue output."""

    id: int = Field(..., alias=SlurmSqueueHeader.JOBID.value)
    step: str = Field(..., alias=SlurmSqueueHeader.NAME.value)
    status: SlurmJobStatus = Field(..., alias=SlurmSqueueHeader.STATE.value)
    time_limit: str = Field(..., alias=SlurmSqueueHeader.TIME_LIMIT.value)
    time_elapsed: int = Field(..., alias=SlurmSqueueHeader.TIME.value)
    started_at: Optional[datetime] = Field(None, alias=SlurmSqueueHeader.START_TIME.value)

    @field_validator("status", mode="before")
    @classmethod
    def convert_status_to_lower_case(cls, raw_status: str) -> str:
        """Convert string to lower case."""
        return raw_status.lower()

    @field_validator("time_elapsed", mode="before")
    @classmethod
    def convert_time_elapsed_to_minutes(cls, raw_time_elapsed: str) -> int:
        """Convert squeue timestamp string into minutes."""
        if not raw_time_elapsed or not isinstance(raw_time_elapsed, str):
            return 0
        timestamps: List[str] = []
        if "-" in raw_time_elapsed:
            timestamps: List[str] = raw_time_elapsed.split("-")
        day_nr: int = int(timestamps[0]) if timestamps else 0
        raw_timestamp: str = timestamps[1] if timestamps else raw_time_elapsed
        time_elapsed: datetime = get_datetime_from_timestamp(
            timestamp=raw_timestamp, datetime_formats=["%M:%S", "%H:%M:%S"]
        )
        return convert_timestamp_to_minutes(timestamp=time_elapsed) + convert_days_to_minutes(
            days_nr=day_nr
        )

    @field_validator("started_at", mode="before")
    @classmethod
    def convert_started_at_to_datetime(cls, raw_started: str) -> Optional[datetime]:
        """Convert started to datetime if string is in datetime format."""
        if isinstance(raw_started, str) and raw_started not in {"N/A"}:
            return datetime.strptime(raw_started, "%Y-%m-%dT%H:%M:%S")


class SqueueResult(BaseModel):
    """Model used to parse SLURM squeue output."""

    jobs: List[SqueueJob]
    jobs_status_distribution: Optional[Dict[str, float]] = None

    @model_validator(mode="after")
    def set_jobs_status_distribution(self):
        """Set the fraction for each status present in the squeue jobs.
        Example {PENDING:0.33, RUNNING:0.33, FAILED: 0.33}."""
        if not self.jobs:
            raise MissingSqueueOutput("No squeue output")
        job_statuses: List[str] = [job.status for job in self.jobs]
        self.jobs_status_distribution = {
            job_status: round(job_statuses.count(job_status) / len(self.jobs), 2)
            for job_status in set(job_statuses)
        }



from typing import Dict, Optional

import pytest

from trailblazer.apps.slurm.api import (
    get_current_analysis_status,
    get_squeue_result,
    reformat_squeue_result_job_step,
)
from trailblazer.apps.slurm.models import SqueueResult
from trailblazer.constants import Pipeline, SlurmJobStatus, TrailblazerStatus


def test_get_squeue_result(squeue_stream_jobs):
    # GIVEN a squeue stream

    # WHEN getting the squeue results
    squeue_result: SqueueResult = get_squeue_result(squeue_response=squeue_stream_jobs)

    # THEN it should return squeue results and jobs
    assert isinstance(squeue_result, SqueueResult)
    assert isinstance(squeue_result.jobs, list)


@pytest.mark.parametrize(
    "data_analysis, raw_job_step, expected_job_step",
    [
        (Pipeline.MIP_DNA, "mipdnastep_jobstep", "mipdnastep"),
        (Pipeline.MIP_RNA, "miprnastep_jobstep", "miprnastep"),
        (Pipeline.BALSAMIC, "job.step.balsamicstep", "balsamicstep"),
        (Pipeline.SARS_COV_2, "job_step_mutantstep", "mutantstep"),
    ],
)
def test_reformat_squeue_result_job_step(
    data_analysis: str, raw_job_step: str, expected_job_step: str
):
    """Test reformatting job step name for each pipeline."""
    # GIVEN a data analysis and a job step

    # WHEN reformation the job step
    reformatted_job_step: str = reformat_squeue_result_job_step(
        data_analysis=data_analysis, job_step=raw_job_step
    )

    # THEN it should return a reformatted job step
    assert reformatted_job_step == expected_job_step


@pytest.mark.parametrize(
    "data_analysis, raw_job_step, expected_job_step",
    [
        (Pipeline.MIP_DNA, "jobstep", "jobstep"),
        (Pipeline.MIP_RNA, "jobstep", "jobstep"),
        (Pipeline.BALSAMIC, "job.step", "job.step"),
        (Pipeline.SARS_COV_2, "jobstep", "jobstep"),
    ],
)
def test_reformat_squeue_result_job_step_malformed_job_step(
    caplog, data_analysis: str, raw_job_step: str, expected_job_step: str
):
    """Test reformatting job step name for each pipeline when job step is malformed."""

    caplog.set_level("ERROR")
    # GIVEN a data analysis and a job step

    # WHEN reformation the job step
    reformatted_job_step: str = reformat_squeue_result_job_step(
        data_analysis=data_analysis, job_step=raw_job_step
    )

    # THEN it should return the original job step
    assert reformatted_job_step == expected_job_step
    assert f"Job step - {raw_job_step}: is malformed" in caplog.text


@pytest.mark.parametrize(
    "analysis_status, job_status_distribution, expected_analysis_status",
    [
        ("failed", {SlurmJobStatus.FAILED: 0.01}, TrailblazerStatus.FAILED),
        (
            "failed_ongoing",
            {SlurmJobStatus.FAILED: 0.01, SlurmJobStatus.RUNNING: 0.1},
            TrailblazerStatus.ERROR,
        ),
        ("error", {SlurmJobStatus.TIME_OUT: 0.01}, TrailblazerStatus.FAILED),
        (
            "error_ongoing",
            {SlurmJobStatus.TIME_OUT: 0.01, SlurmJobStatus.PENDING: 0.1},
            TrailblazerStatus.ERROR,
        ),
        ("completed", {SlurmJobStatus.COMPLETED: 1.0}, TrailblazerStatus.COMPLETED),
        ("pending", {SlurmJobStatus.PENDING: 1.0}, TrailblazerStatus.PENDING),
        (
            "running",
            {SlurmJobStatus.RUNNING: 0.01, SlurmJobStatus.PENDING: 0.1},
            TrailblazerStatus.RUNNING,
        ),
        ("canceled", {SlurmJobStatus.CANCELLED: 0.01}, TrailblazerStatus.CANCELLED),
        (
            "canceled_ongoing",
            {SlurmJobStatus.CANCELLED: 0.01, SlurmJobStatus.RUNNING: 0.1},
            TrailblazerStatus.RUNNING,
        ),
        (
            "canceled_ongoing_round_to_zero",
            {SlurmJobStatus.CANCELLED: 0.01, SlurmJobStatus.RUNNING: 0.00},
            TrailblazerStatus.RUNNING,
        ),
    ],
)
def test_get_current_analysis_status(
    analysis_status: str, job_status_distribution: Dict[str, float], expected_analysis_status: str
):
    """Test return current analysis status from jobs status distribution."""
    # GIVEN a data analysis and a job status distribution

    # WHEN reformation the job step
    current_analysis_status: Optional[str] = get_current_analysis_status(
        jobs_status_distribution=job_status_distribution
    )

    # THEN it should return the current job status distribution
    assert current_analysis_status == expected_analysis_status



from datetime import datetime
from typing import Dict

import pytest
from click.testing import CliRunner

import trailblazer
from tests.mocks.store_mock import MockStore
from trailblazer.cli.core import (
    add_user_to_db,
    archive_user,
    base,
    cancel,
    delete,
    get_user_from_db,
    get_users_from_db,
    ls_cmd,
    scan,
    set_analysis_completed,
    set_analysis_status,
    unarchive_user,
)
from trailblazer.constants import SlurmJobStatus, TrailblazerStatus
from trailblazer.store.api import Store
from trailblazer.store.models import Analysis


def test_base(cli_runner):
    # GIVEN the cli

    # WHEN asking for the version
    result = cli_runner.invoke(base, ["--version"])

    # THEN it should print the name and version of the tool only
    assert trailblazer.__title__ in result.output
    assert trailblazer.__version__ in result.output


def test_set_analysis_completed(cli_runner, trailblazer_context, caplog):
    # GIVEN an analysis with status FAILED
    failed_analysis = "crackpanda"
    analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(case_id=failed_analysis)

    # Make sure status is not "completed"
    assert analysis_obj.status != "completed"

    # WHEN running command
    result = cli_runner.invoke(
        set_analysis_completed, [str(analysis_obj.id)], obj=trailblazer_context
    )

    # THEN command runs successfully
    assert result.exit_code == 0

    # THEN status will be set to COMPLETED
    analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(case_id=failed_analysis)
    assert analysis_obj.status == "completed"


def test_set_analysis_status(cli_runner, trailblazer_context, caplog) -> None:
    """Test that the lastest analysis status is updated for a given case ID."""

    # GIVEN an analysis with status FAILED
    failed_analysis = "crackpanda"
    analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(case_id=failed_analysis)

    # Make sure status is not "completed"
    assert analysis_obj.status != "qc"

    # WHEN running command
    result = cli_runner.invoke(
        set_analysis_status, ["--status", "qc", failed_analysis], obj=trailblazer_context
    )

    # THEN command runs successfully
    assert result.exit_code == 0

    # THEN status will be set to COMPLETED
    analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(case_id=failed_analysis)
    assert analysis_obj.status == "qc"


def test_set_analysis_status_error(cli_runner, trailblazer_context, caplog) -> None:
    """Test that setting the status to a non-accepted value raises an error."""

    # GIVEN an analysis with status FAILED
    failed_analysis = "crackpanda"

    # WHEN running command
    result = cli_runner.invoke(
        set_analysis_status,
        ["--status", "non_existing_status", failed_analysis],
        obj=trailblazer_context,
    )

    # THEN an error should be raised
    assert result.exit_code != 0
    assert "Invalid status" in caplog.text


def test_cancel_nonexistent(cli_runner, trailblazer_context, caplog):
    with caplog.at_level("ERROR"):
        # GIVEN Trailblazer database with analyses and jobs
        trailblazer_context["trailblazer"].update_ongoing_analyses()

        # WHEN running cancel on non-existing entry
        result = cli_runner.invoke(cancel, ["123"], obj=trailblazer_context)

        # THEN command runs successfully
        assert result.exit_code == 0

        # THEN error log informs user that analysis does not exist
        assert "does not exist" in caplog.text


def test_cancel_not_running(cli_runner, trailblazer_context, caplog):
    with caplog.at_level("ERROR"):
        # GIVEN an analysis that is NOT running
        failed_analysis = "crackpanda"
        analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(
            case_id=failed_analysis
        )
        # WHEN trying to cancel analysis
        trailblazer_context["trailblazer"].update_ongoing_analyses()
        result = cli_runner.invoke(cancel, [str(analysis_obj.id)], obj=trailblazer_context)

        # THEN command runs successfully
        assert result.exit_code == 0

        # THEN error log informs user that analysis cannot be cancelled as it is not running
        assert "is not running" in caplog.text


def test_cancel_ongoing_analysis(cli_runner, trailblazer_context, caplog):
    """Test all ongoing analysis jobs are cancelled."""
    caplog.set_level("INFO")

    # GIVEN an ongoing analysis
    trailblazer_db: Store = trailblazer_context["trailblazer"]
    trailblazer_db.update_ongoing_analyses()
    analysis: Analysis = trailblazer_db.get_latest_analysis(case_id="blazinginsect")

    # Analysis should have jobs that can be cancelled
    assert analysis.jobs

    # WHEN running cancel command
    result = cli_runner.invoke(cancel, [str(analysis.id)], obj=trailblazer_context)

    # THEN command should run successfully
    assert result.exit_code == 0

    # THEN log should inform of successful cancellation
    assert "all ongoing jobs cancelled successfully" in caplog.text
    assert "Cancelling" in caplog.text

    # THEN job id from squeue output will be cancelled
    assert f"{analysis.jobs[3].slurm_id}" in caplog.text

    # THEN analysis status is set to cancelled
    assert SlurmJobStatus.CANCELLED in analysis.comment
    assert analysis.status == TrailblazerStatus.CANCELLED


def test_delete_nonexisting(cli_runner, trailblazer_context, caplog):
    with caplog.at_level("ERROR"):
        # GIVEN Trailblazer database with analyses and jobs
        trailblazer_context["trailblazer"].update_ongoing_analyses()

        # WHEN trying to delete analysis not in database
        cli_runner.invoke(delete, ["123"], obj=trailblazer_context)

        # THEN error log informs user that analysis is not found
        assert "Analysis not found" in caplog.text


def test_delete_ongoing_fail(cli_runner, trailblazer_context, caplog):
    with caplog.at_level("ERROR"):
        # GIVEN Trailblazer database with analyses and jobs
        trailblazer_context["trailblazer"].update_ongoing_analyses()

        # GIVEN an analysis that is ongoing
        analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(case_id="escapedgoat")

        # WHEN trying to delete ongoing analysis without --force flag
        result = cli_runner.invoke(delete, [str(analysis_obj.id)], obj=trailblazer_context)

        # THEN command executes successfully
        assert result.exit_code == 0

        # THEN error log informs user that ongoing analysis cannot be deleted without --force flag
        assert "--force" in caplog.text

        # THEN analysis should not be cancelled
        assert analysis_obj.status != "canceled"


def test_delete_ongoing_force(cli_runner, trailblazer_context, caplog):
    with caplog.at_level("INFO"):
        # GIVEN Trailblazer database with analyses and jobs and an ongoing analysis
        trailblazer_context["trailblazer"].update_ongoing_analyses()
        analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(case_id="escapedgoat")

        # WHEN running command with --force flag
        result = cli_runner.invoke(
            delete, [str(analysis_obj.id), "--force"], obj=trailblazer_context
        )

        # THEN command runs successfully
        assert result.exit_code == 0

        # THEN log informs user that Trailblazer is deleting analysis
        assert "Deleting" in caplog.text


def test_get_user_not_in_database(cli_runner, trailblazer_context: Dict[str, MockStore], caplog):
    """Test getting user when user is not in the database."""
    # GIVEN populated Trailblazer database

    caplog.set_level("ERROR")

    # WHEN looking for a user that is not added
    cli_runner.invoke(get_user_from_db, ["not_in_db@disco.com"], obj=trailblazer_context)

    # THEN log should inform that user was not found
    assert "not found" in caplog.text


def test_get_user_in_database(
    cli_runner, trailblazer_context: Dict[str, MockStore], caplog, user_email: str, username: str
):
    """Test getting user when user is in the database."""
    # GIVEN populated Trailblazer database
    caplog.set_level("INFO")

    # WHEN looking for a user that is added
    cli_runner.invoke(get_user_from_db, [user_email], obj=trailblazer_context)

    # THEN log informs the user about their full name
    assert username in caplog.text


def test_add_user(cli_runner, trailblazer_context: Dict[str, MockStore], caplog):
    """Test adding a user to the database."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN adding new user
    cli_runner.invoke(
        add_user_to_db, ["harriers@disco.com", "--name", "Harry DuBois"], obj=trailblazer_context
    )

    # THEN log should inform that user is added
    assert "New user added" in caplog.text


def test_add_user_when_already_exists(
    cli_runner, trailblazer_context: Dict[str, MockStore], caplog, user_email: str, username: str
):
    """Test adding a user to the database when the user already exists."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN adding new user
    cli_runner.invoke(add_user_to_db, [user_email, "--name", username], obj=trailblazer_context)

    # THEN log should inform that user is added
    assert f"User with email {user_email} found" in caplog.text


def test_users(
    cli_runner, trailblazer_context: Dict[str, MockStore], caplog, user_email: str, username: str
):
    """Test getting users from the database."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN getting users
    cli_runner.invoke(get_users_from_db, ["--email", user_email], obj=trailblazer_context)

    # THEN log shows users
    assert username in caplog.text


def test_archive_user(
    cli_runner, trailblazer_context: Dict[str, MockStore], caplog, user_email: str
):
    """Test archiving a user in the database."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN archiving user
    cli_runner.invoke(archive_user, [user_email], obj=trailblazer_context)

    # THEN log shows users as archived
    assert f"User archived: {user_email}" in caplog.text


def test_archive_user_when_already_archived(
    cli_runner, trailblazer_context: Dict[str, MockStore], caplog, archived_user_email: str
):
    """Test archiving a user in the database when already archived."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN archiving user
    cli_runner.invoke(archive_user, [archived_user_email], obj=trailblazer_context)

    # THEN log shows users is already archived
    assert "has archive status: True" in caplog.text


def test_archive_user_when_non_existing_user(
    cli_runner, trailblazer_context: Dict[str, MockStore], caplog, archived_user_email: str
):
    """Test archiving a user in the database when user does not exist."""
    # GIVEN populated Trailblazer database

    # GIVEN a user email that does not exist
    non_existing_email: str = "does-not.exist@none.com"

    # WHEN archiving user
    cli_runner.invoke(archive_user, [non_existing_email], obj=trailblazer_context)

    # THEN log shows users is not found
    assert f"User with email {non_existing_email} not found" in caplog.text


def test_unarchive_user_when_not_archived(
    cli_runner, trailblazer_context: Dict[str, MockStore], caplog, user_email: str
):
    """Test unarchiving a user in the database which is not archived."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN unarchiving user
    cli_runner.invoke(unarchive_user, [user_email], obj=trailblazer_context)

    # THEN log shows users is not archived
    assert f"User with email {user_email} has archive status: False" in caplog.text


def test_unarchive_user(
    cli_runner, trailblazer_context: Dict[str, MockStore], caplog, archived_user_email: str
):
    """Test unarchiving a user in the database which is archived."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN unarchiving user
    cli_runner.invoke(unarchive_user, [archived_user_email], obj=trailblazer_context)

    # THEN log shows users as not archived
    assert f"User unarchived: {archived_user_email}" in caplog.text


def test_scan(cli_runner, trailblazer_context, caplog):
    with caplog.at_level("INFO"):
        # GIVEN populated Trailblazer database with pending analyses

        # GIVEN an analysis that is pending
        analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(
            case_id="blazinginsect"
        )
        assert analysis_obj.status == "pending"

        # WHEN running trailblazer scan command
        cli_runner.invoke(scan, [], obj=trailblazer_context)
        assert "All analyses updated" in caplog.text
        analysis_obj = trailblazer_context["trailblazer"].get_latest_analysis(
            case_id="blazinginsect"
        )

        # THEN the stats of analysis should be updated as expected
        assert analysis_obj.status == "running"


@pytest.mark.parametrize(
    "case_id, status",
    [
        ("blazinginsect", "running"),
        ("crackpanda", "failed"),
        ("fancymole", "completed"),
        ("happycow", "pending"),
    ],
)
def test_ls(
    cli_runner: CliRunner,
    process_exit_success: int,
    trailblazer_context: Dict[str, MockStore],
    case_id: str,
    status: str,
    timestamp_now: datetime,
):
    """Test the Traiblazer ls CLI command using different cases and statuses."""
    # GIVEN populated Trailblazer database with pending analyses
    trailblazer_db: Store = trailblazer_context["trailblazer"]

    # Update analyses to their expected status
    trailblazer_db.update_ongoing_analyses()

    # WHEN running ls command while filtering by status
    result = cli_runner.invoke(
        ls_cmd, ["--status", status, "--before", str(timestamp_now.date())], obj=trailblazer_context
    )

    # THEN command runs successfully with provided options
    assert result.exit_code == process_exit_success

    # THEN ls print info about cases with that status
    assert case_id in result.output



from typing import List

from tests.mocks.store_mock import MockStore
from trailblazer.apps.slurm.api import get_squeue_result
from trailblazer.apps.slurm.models import SqueueResult
from trailblazer.constants import TrailblazerStatus
from trailblazer.store.filters.user_filters import UserFilter, apply_user_filter
from trailblazer.store.models import Analysis, User


def test_update_analysis_jobs(analysis_store: MockStore, tower_jobs: List[dict], case_id: str):
    """Test jobs are successfully updated."""

    # GIVEN an analysis without failed jobs
    analysis: Analysis = analysis_store.get_latest_analysis(case_id=case_id)
    assert not analysis.jobs

    # WHEN jobs are updated
    analysis_store.update_analysis_jobs(analysis=analysis, jobs=tower_jobs[:2])

    # THEN there should be jobs
    assert analysis.jobs


def test_update_user_is_archived(user_store: MockStore, user_email: str):
    """Test updating user is archived attribute."""
    # GIVE a database and a not archived user
    user: User = apply_user_filter(
        filter_functions=[UserFilter.FILTER_BY_EMAIL],
        users=user_store.get_query(table=User),
        email=user_email,
    ).first()
    assert not user.is_archived

    # WHEN updating a user archive status
    user_store.update_user_is_archived(user=user, archive=True)

    archived_user: User = apply_user_filter(
        filter_functions=[UserFilter.FILTER_BY_EMAIL],
        users=user_store.get_query(table=User),
        email=user_email,
    ).first()

    # THEN user should be recorded as archived in the database
    assert archived_user


def test_update_analysis_jobs_from_slurm_jobs(analysis_store: MockStore, squeue_stream_jobs: str):
    """Test updating analysis jobs when given squeue results."""
    # GIVEN an analysis and a squeue stream
    analysis: Analysis = analysis_store.get_query(table=Analysis).first()
    assert not analysis.jobs

    squeue_result: SqueueResult = get_squeue_result(squeue_response=squeue_stream_jobs)

    # WHEN updating the analysis
    analysis_store.update_analysis_jobs_from_slurm_jobs(
        analysis=analysis, squeue_result=squeue_result
    )
    updated_analysis: Analysis = analysis_store.get_analysis(
        case_name=analysis.family, started_at=analysis.started_at, status=TrailblazerStatus.PENDING
    )

    # THEN it should update the analysis jobs
    assert updated_analysis.jobs



from datetime import datetime
from typing import Callable, Dict, List, Optional, Union

from trailblazer.store.base import BaseHandler_2
from trailblazer.store.filters.analyses_filters import (
    AnalysisFilter,
    apply_analysis_filter,
)
from trailblazer.store.filters.job_filters import JobFilter, apply_job_filter
from trailblazer.store.filters.user_filters import UserFilter, apply_user_filter
from trailblazer.store.models import Analysis, Job, User


class ReadHandler(BaseHandler_2):
    """Class for reading items in the database."""

    def get_nr_jobs_with_status_per_category(
        self, status: str, since_when: datetime = None
    ) -> List[Dict[str, Union[str, int]]]:
        """Return the number of jobs with status per category (name)."""
        filter_map: Dict[Callable, Optional[Union[str, bool]]] = {
            JobFilter.FILTER_BY_STATUS: status,
            JobFilter.FILTER_BY_SINCE_WHEN: since_when,
        }
        filter_functions: List[Callable] = [
            function for function, supplied_arg in filter_map.items() if supplied_arg
        ]
        categories = apply_job_filter(
            filter_functions=filter_functions,
            jobs=self.get_job_query_with_name_and_count_labels(),
            since_when=since_when,
            status=status,
        )
        categories = categories.group_by(Job.name).all()
        return [{"name": category.name, "count": category.count} for category in categories]

    def get_analysis(self, case_name: str, started_at: datetime, status: str) -> Optional[Analysis]:
        """Return the latest analysis for supplied parameters."""
        return apply_analysis_filter(
            filter_functions=[
                AnalysisFilter.FILTER_BY_CASE_NAME,
                AnalysisFilter.FILTER_BY_STARTED_AT,
                AnalysisFilter.FILTER_BY_STATUS,
            ],
            analyses=self.get_query(table=Analysis),
            case_name=case_name,
            started_at=started_at,
            status=status,
        ).first()

    def get_analysis_with_id(self, analysis_id: int) -> Optional[Analysis]:
        """Get a single analysis by id."""
        return apply_analysis_filter(
            filter_functions=[AnalysisFilter.FILTER_BY_ID],
            analyses=self.get_query(table=Analysis),
            analysis_id=analysis_id,
        ).first()

    def get_user(
        self,
        email: str = None,
        exclude_archived: bool = True,
    ) -> User:
        """Return user from the database."""
        filter_map: Dict[Callable, Optional[Union[str, bool]]] = {
            UserFilter.FILTER_BY_CONTAINS_EMAIL: email,
            UserFilter.FILTER_BY_IS_NOT_ARCHIVED: exclude_archived,
        }
        filter_functions: List[Callable] = [
            function for function, supplied_arg in filter_map.items() if supplied_arg
        ]
        return apply_user_filter(
            filter_functions=filter_functions,
            users=self.get_query(table=User),
            email=email,
        ).first()

    def get_users(
        self,
        name: str = None,
        email: str = None,
        exclude_archived: bool = True,
    ) -> List[User]:
        """Return users from the database."""
        filter_map: Dict[Callable, Optional[Union[str, bool]]] = {
            UserFilter.FILTER_BY_CONTAINS_EMAIL: email,
            UserFilter.FILTER_BY_CONTAINS_NAME: name,
            UserFilter.FILTER_BY_IS_NOT_ARCHIVED: exclude_archived,
        }
        filter_functions: List[Callable] = [
            function for function, supplied_arg in filter_map.items() if supplied_arg
        ]
        return apply_user_filter(
            filter_functions=filter_functions,
            users=self.get_query(table=User),
            email=email,
            name=name,
        ).all()



from trailblazer.store.crud.create import CreateHandler
from trailblazer.store.crud.delete import DeleteHandler
from trailblazer.store.crud.read import ReadHandler
from trailblazer.store.crud.update import UpdateHandler


class CoreHandler(
    CreateHandler,
    DeleteHandler,
    ReadHandler,
    UpdateHandler,
):
    """Aggregating class for the store API handlers."""

    def __init__(self):
        pass



import pytest

from datetime import datetime

from trailblazer.utils.datetime import (
    get_date_number_of_days_ago,
    convert_days_to_minutes,
    convert_timestamp_to_minutes,
    get_datetime_from_timestamp,
)

NUMBER_OF_MINUTES_IN_TWO_DAYS: int = 2880
MAX_NR_MINUTES_IN_TIMESTAMP: int = 1439


@pytest.mark.parametrize(
    "nr_of_days, expected_minutes",
    [(0, 0), (2, NUMBER_OF_MINUTES_IN_TWO_DAYS)],
)
def test_convert_days_to_minutes(nr_of_days: int, expected_minutes: int):
    """Test converting days to minutes."""

    # GIVEN days

    # WHEN calling the function
    minutes: int = convert_days_to_minutes(days_nr=nr_of_days)

    # THEN the return is number of minutes corresponding to the number of days
    assert minutes == expected_minutes


@pytest.mark.parametrize(
    "timestamp, expected_minutes",
    [
        (datetime.strptime("0:0:0", "%H:%M:%S"), 0),
        (datetime.strptime("23:59:59", "%H:%M:%S"), MAX_NR_MINUTES_IN_TIMESTAMP),
    ],
)
def test_convert_timestamp_to_minutes(timestamp: datetime, expected_minutes: int):
    """Test converting timestamp to minutes."""

    # GIVEN timestamps

    # WHEN calling the function
    minutes: int = convert_timestamp_to_minutes(timestamp=timestamp)

    # THEN the return is number of minutes corresponding to timestamp
    assert minutes == expected_minutes


def test_get_datetime_from_timestamp():
    """Test returning a datetime object from a timestamp."""

    # GIVEN a timestamp

    # WHEN calling the function
    date: datetime = get_datetime_from_timestamp(
        timestamp="23:59:59", datetime_formats=["%H:%M:%S"]
    )

    # THEN the return should be a datetime object
    assert isinstance(date, datetime)


def test_get_datetime_from_timestamp_with_invalid_timestamp(caplog):
    """Test returning a datetime object from a timestamp, which is invalid."""

    # GIVEN a timestamp

    # WHEN calling the function
    get_datetime_from_timestamp(timestamp="not a timestamp", datetime_formats=["%H:%M:%S"])

    # THEN log the error
    assert "Error converting timestamp" in caplog.text


def test_get_date_number_of_days_ago(timestamp_now: datetime):
    """Test returning a date corresponding to number of days ago."""

    # GIVEN days ago

    # WHEN calling the function
    date: datetime = get_date_number_of_days_ago(number_of_days_ago=1)

    # THEN the return should be a date
    assert isinstance(date, datetime)

    # Then the date returned should be less than then today
    assert date < timestamp_now



from sqlalchemy.orm import Query

from tests.mocks.store_mock import MockStore
from trailblazer.store.filters.analyses_filters import (
    filter_analyses_by_case_name,
    filter_analyses_by_id,
    filter_analyses_by_started_at,
    filter_analyses_by_status,
)
from trailblazer.store.models import Analysis


def test_filter_analyses_by_id(analysis_store: MockStore):
    """Test return analysis by id when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by id
    analysis: Query = filter_analyses_by_id(
        analyses=analysis_store.get_query(table=Analysis), analysis_id=existing_analysis.id
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()


def test_filter_analyses_by_case_name(analysis_store: MockStore):
    """Test return analysis by case name when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by case name
    analysis: Query = filter_analyses_by_case_name(
        analyses=analysis_store.get_query(table=Analysis), case_name=existing_analysis.family
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()


def test_filter_analyses_by_started_at(analysis_store: MockStore):
    """Test return analysis by started at when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by when it was started
    analysis: Query = filter_analyses_by_started_at(
        analyses=analysis_store.get_query(table=Analysis), started_at=existing_analysis.started_at
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()


def test_filter_analyses_by_status(analysis_store: MockStore):
    """Test return analysis by status when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by status
    analysis: Query = filter_analyses_by_status(
        analyses=analysis_store.get_query(table=Analysis), status=existing_analysis.status
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()
